from functools import lru_cache

from django import forms

from . import versionables


class VersionContentChoiceField(forms.ModelChoiceField):
    """Form field used to display a list of grouper instances"""

    def __init__(self, *args, **kwargs):
        self.language = kwargs.pop("language")
        self.predefined_label_method = kwargs.pop("option_label_override")
        super().__init__(*args, **kwargs)

    def label_from_instance(self, obj):
        """Overridden to allow customizing the labels of the groupers"""
        if self.predefined_label_method:
            return self.predefined_label_method(obj, self.language)
        else:
            return super().label_from_instance(obj)


@lru_cache
def grouper_form_factory(content_model, language=None):
    """Returns a form class used for selecting a grouper to see versions of.
    Form has a single field - grouper - which is a model choice field
    with available grouper objects for specified content model.

    :param content_model: Content model class
    :param language: Language
    """
    versionable = versionables.for_content(content_model)
    valid_grouper_pk = content_model.admin_manager\
        .latest_content()\
        .values_list(versionable.grouper_field_name, flat=True)

    return type(
        content_model.__name__ + "GrouperForm",
        (forms.Form,),
        {
            "_content_model": content_model,
            versionable.grouper_field_name: VersionContentChoiceField(
                queryset=versionable.grouper_model.objects.filter(
                    pk__in=valid_grouper_pk,
                ),
                label=versionable.grouper_model._meta.verbose_name,
                option_label_override=versionable.grouper_selector_option_label,
                language=language,
            ),
        },
    )



from itertools import chain

from django.contrib.contenttypes.models import ContentType
from django.db.models import Max, Prefetch
from django.utils.functional import cached_property

from .admin import VersioningAdminMixin
from .helpers import get_content_types_with_subclasses
from .models import Version


class BaseVersionableItem:
    concrete = False

    def __init__(self, content_model, content_admin_mixin=None):
        self.content_model = content_model
        self.content_admin_mixin = content_admin_mixin or VersioningAdminMixin


class VersionableItem(BaseVersionableItem):
    concrete = True

    def __init__(
        self,
        content_model,
        grouper_field_name,
        copy_function,
        extra_grouping_fields=None,
        version_list_filter_lookups=None,
        on_publish=None,
        on_unpublish=None,
        on_draft_create=None,
        on_archive=None,
        grouper_selector_option_label=False,
        content_admin_mixin=None,
        preview_url=None,
    ):
        super().__init__(content_model, content_admin_mixin)
        # Set the grouper field
        self.grouper_field_name = grouper_field_name
        self.grouper_field = self._get_grouper_field()
        self.extra_grouping_fields = extra_grouping_fields or ()
        # Set the grouper selector label
        self.grouper_selector_option_label = grouper_selector_option_label
        self.version_list_filter_lookups = version_list_filter_lookups or {}
        self.copy_function = copy_function
        self.on_publish = on_publish
        self.on_unpublish = on_unpublish
        self.on_draft_create = on_draft_create
        self.on_archive = on_archive
        self.preview_url = preview_url

    def _get_grouper_field(self):
        """Get the grouper field on the content model

        :return: instance of a django model field
        """
        return self.content_model._meta.get_field(self.grouper_field_name)

    @cached_property
    def version_model_proxy(self):
        """Returns a dynamically created proxy model class to Version.
        It's used for creating separate version model classes for each
        content type.
        """
        model_name = self.content_model.__name__ + "Version"

        ProxyVersion = type(
            model_name,
            (Version,),
            {
                "Meta": type("Meta", (), {"proxy": True, "managed": False}),
                "__module__": __name__,
                "_source_model": self.content_model,
            },
        )
        return ProxyVersion

    @property
    def grouper_model(self):
        """Returns the grouper model class"""
        return self.grouper_field.remote_field.model

    @cached_property
    def content_model_is_sideframe_editable(self):
        """Determine if a content model can be opened in the sideframe or not.

        :return: Default True, False if the content model is not suitable for the sideframe
        :rtype: bool
        """
        from cms.models import Placeholder

        # Any models that contain a placeholder are deemed as not sideframe editing
        # compatible i.e. they are toolbar enabled and contain placeholders. We can't
        # use toolbar enabled status alone because any models that use version compare
        # enable the toolbar to display.
        for field in self.content_model._meta.get_fields(include_hidden=True):
            if field.related_model == Placeholder:
                return False
        return True

    def distinct_groupers(self, **kwargs):
        """Returns a queryset of `self.content` objects with unique
        grouper objects.

        Useful for listing, e.g. all Polls.

        :param kwargs: Optional filtering parameters for inner queryset
        """
        queryset = self.content_model._base_manager.values(
            self.grouper_field.name
        ).filter(**kwargs)
        inner = queryset.annotate(Max("pk")).values("pk__max")
        return self.content_model._base_manager.filter(id__in=inner)

    def for_grouper(self, grouper):
        """Returns all `Content` objects for specified grouper object."""
        return self.for_grouping_values(**{self.grouper_field.name: grouper})

    def for_content_grouping_values(self, content):
        """Returns all `Content` objects based on all grouping values
        in specified content object."""
        return self.for_grouping_values(**self.grouping_values(content))

    def for_grouping_values(self, **kwargs):
        """Returns all `Content` objects based on all specified
        grouping values."""
        return self.content_model._base_manager.filter(**kwargs)

    @property
    def grouping_fields(self):
        """Returns an iterator for all the grouping fields"""
        return chain([self.grouper_field_name], self.extra_grouping_fields)

    def grouping_values(self, content, relation_suffix=True):
        """Returns a dict of grouper fields as keys and values from the content instance

        :param content: instance of a content model
        :param relation_suffix: bool setting whether fk fieldnames have '_id' added
        :return: a dict like {'grouping_field1': content.grouping_field1, ...}
        """
        def suffix(field, allow=True):
            if allow and content._meta.get_field(field).is_relation:
                return field + "_id"
            return field

        return {
            suffix(field, allow=relation_suffix): getattr(content, suffix(field))
            for field in self.grouping_fields
        }

    def grouper_choices_queryset(self):
        """Returns a queryset of all the available groupers instances of the registered type"""
        content_objects = self.content_model.admin_manager.all().latest_content()
        cache_name = self.grouper_field.remote_field.get_accessor_name()
        return self.grouper_model._base_manager.prefetch_related(
            Prefetch(cache_name, queryset=content_objects)
        )

    def get_grouper_with_fallbacks(self, grouper_id):
        return self.grouper_choices_queryset().filter(pk=grouper_id).first()

    def _get_content_types(self):
        return [ContentType.objects.get_for_model(self.content_model).pk]

    @cached_property
    def content_types(self):
        """Get the primary key of the content type of the registered content model.

        :return:  A list with the primary keys of the content types
        """
        # NOTE: If using this class this will be a list with one element,
        # but PolymorphicVersionableItem overrides this and can return
        # more elements
        return self._get_content_types()


class PolymorphicVersionableItem(VersionableItem):
    """VersionableItem for use by base polymorphic class
    (for example filer.File).
    """

    def _get_content_types(self):
        return get_content_types_with_subclasses([self.content_model])


class VersionableItemAlias(BaseVersionableItem):
    """VersionableItem that points to a different VersionableItem,
    so that all operations are executed in context of
    the other VersionableItem.
    """

    def __init__(self, content_model, to, content_admin_mixin=None):
        super().__init__(content_model, content_admin_mixin)
        self.to = to

    def __getattr__(self, name):
        return getattr(self.to, name)


def default_copy(original_content):
    """Copy all fields of the original content object exactly as they are
    and return a new content object which is different only in its pk.

    NOTE: This will only work for very simple content objects. This will
    throw exceptions on one2one and m2m relationships. And it might not
    be the desired behaviour for some foreign keys (in some cases we
    would expect a version to copy some of its related objects as well).
    In such cases a custom copy method must be defined and specified in
    cms_config.py

    NOTE: A custom copy method will need to use the content model's
    _original_manage to create only a content model object and not also a Version object.
    """
    content_model = original_content.__class__
    content_fields = {
        field.name: getattr(original_content, field.name)
        for field in content_model._meta.fields
        # don't copy primary key because we're creating a new obj
        if content_model._meta.pk.name != field.name
    }
    # Use original manager to avoid creating a new draft version here!
    return content_model._original_manager.create(**content_fields)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.apps import apps
from django.db import migrations
from django.db.models import Max


def _distinct_groupers(versionable):
    inner = (
        versionable.content_model._base_manager.values(*versionable.grouping_fields)
        .annotate(Max("pk"))
        .values("pk__max")
    )
    return versionable.content_model._base_manager.filter(id__in=inner).all()


def forwards(app_registry, schema_editor):
    """
    Assign version numbers to existing data

    Removed because versions will always be created after this point, and it is causing dependency problems with other packages tests.
    Anyone that requires this migration can install version 0.0.23, which is the last version with it enabled.
    """
    return
    # versioning_extension = apps.get_app_config("djangocms_versioning").cms_extension
    # Version = app_registry.get_model("djangocms_versioning", "Version")
    # ContentType = apps.get_model("contenttypes", "ContentType")

    # # For each registered version model
    # for versionable in versioning_extension.versionables:
    #     # For each distinct grouper, unique to the grouping field values
    #     for content in _distinct_groupers(versionable):

    #         fields = versionable.grouping_values(content)
    #         content_objects = versionable.for_grouping_values(**fields)
    #         content_type = ContentType.objects.get_for_model(versionable.content_model)

    #         version_list = Version.objects.filter(
    #             object_id__in=content_objects, content_type_id=content_type.pk
    #         ).order_by("pk")

    #         previous_number = 0
    #         for version in version_list:
    #             version.number = previous_number + 1
    #             previous_number = version.number
    #             version.save()


class Migration(migrations.Migration):

    dependencies = [("djangocms_versioning", "0011_version_number")]

    operations = [migrations.RunPython(forwards)]



from functools import lru_cache

from cms.plugin_rendering import ContentRenderer, StructureRenderer
from cms.utils.placeholder import rescan_placeholders_for_obj

from . import versionables
from .constants import DRAFT, PUBLISHED


def prefetch_versioned_related_objects(instance, toolbar):
    instance, plugin = instance.get_plugin_instance()

    candidate_fields = [
        f for f in instance._meta.get_fields() if f.is_relation and not f.auto_created
    ]
    for field in candidate_fields:
        try:
            versionable = versionables.for_grouper(field.remote_field.model)
        except KeyError:
            continue
        if toolbar.edit_mode_active or toolbar.preview_mode_active:
            qs = versionable.content_model._base_manager.filter(
                versions__state__in=(DRAFT, PUBLISHED)
            ).order_by("versions__state")
        else:
            qs = versionable.content_model.objects.all()
        related_field = getattr(instance, field.name)
        if related_field:
            filters = {versionable.grouper_field_name: related_field}
            # TODO Figure out grouping values-awareness
            # for extra fields other than hardcoded 'language'
            if "language" in versionable.extra_grouping_fields:
                filters["language"] = toolbar.request_language
            qs = qs.filter(**filters)
            prefetch_cache = {versionable.grouper_field.remote_field.name: qs}
            related_field._prefetched_objects_cache = prefetch_cache


class VersionContentRenderer(ContentRenderer):
    def render_plugin(self, instance, context, placeholder=None, editable=False):
        prefetch_versioned_related_objects(instance, self.toolbar)
        return super().render_plugin(instance, context, placeholder, editable)

    def render_obj_placeholder(
        self, slot, context, inherit, nodelist=None, editable=True
    ):
        # FIXME This is an ad-hoc solution for page-specific rendering
        # code, which by default doesn't work well with versioning.
        # Remove this method once the issue is fixed.
        from cms.models import Placeholder

        current_obj = self.toolbar.get_object()

        # Not page, therefore we will use toolbar object as
        # the current object and render the placeholder
        rescan_placeholders_for_obj(current_obj)
        placeholder = Placeholder.objects.get_for_obj(current_obj).get(slot=slot)
        content = self.render_placeholder(
            placeholder,
            context=context,
            page=current_obj,
            editable=editable,
            use_cache=True,
            nodelist=None,
        )
        return content


class VersionStructureRenderer(StructureRenderer):
    def render_plugin(self, instance, page=None):
        prefetch_versioned_related_objects(instance, self.toolbar)
        return super().render_plugin(instance, page)


class CMSToolbarVersioningMixin:
    @property
    @lru_cache(16)
    def content_renderer(self):
        return VersionContentRenderer(request=self.request)

    @property
    @lru_cache(16)
    def structure_renderer(self):
        return VersionStructureRenderer(request=self.request)



from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.core.management.base import BaseCommand, CommandError

from djangocms_versioning import constants
from djangocms_versioning.conf import DEFAULT_USER, USERNAME_FIELD
from djangocms_versioning.models import Version
from djangocms_versioning.versionables import _cms_extension

User = get_user_model()


class Command(BaseCommand):
    help = 'Creates Version objects for versioned models lacking one. If the DJANGOCMS_VERSIONING_DEFAULT_USER ' \
           'setting is not populated you will have to provide either the --userid or --username option for ' \
           'each Version object needs to be assigned to a user. ' \
           'If multiple content objects for a grouper model are found only the newest (by primary key) is ' \
           'assigned the state, older versions are marked as "archived".'

    def add_arguments(self, parser):
        parser.add_argument(
            "--state",
            type=str,
            default=constants.DRAFT,
            choices=[key for key, value in constants.VERSION_STATES if key != constants.UNPUBLISHED],
            help=f"state of newly created version object (defaults to {constants.DRAFT})"
        )
        parser.add_argument(
            "--username",
            type=str,
            help="Username of user to create the missing Version objects"
        )
        parser.add_argument(
            "--userid",
            type=int,
            help="User id of user to create the missing Version objects"
        )

        parser.add_argument(
            "--dry-run",
            action="store_true",
            help="Do not change the database",
        )

    @staticmethod
    def get_user(options):
        if DEFAULT_USER is not None:  # pragma: no cover
            try:
                return User.objects.get(pk=DEFAULT_USER)
            except User.DoesNotExist as err:
                raise CommandError(
                    f"No user with id {DEFAULT_USER} found "
                    f"(specified as DJANGOCMS_VERSIONING_DEFAULT USER in settings.py"
                ) from err

        if options["userid"] and options["username"]:  # pragma: no cover
            raise CommandError("Only either one of the options '--userid' or '--username' may be given")
        if options["userid"]:
            try:
                return User.objects.get(pk=options["userid"])
            except User.DoesNotExist as err:
                raise CommandError(f"No user with id {options['userid']} found") from err
        if options["username"]:  # pragma: no cover
            try:
                return User.objects.get(**{USERNAME_FIELD: options["username"]})
            except User.DoesNotExist as err:
                raise CommandError(f"No user with name {options['username']} found") from err
        return None  # pragma: no cover

    def handle(self, *args, **options):
        user = self.get_user(options)

        for versionable in _cms_extension().versionables:
            Model = versionable.content_model
            content_type = ContentType.objects.get_for_model(Model)
            version_ids = Version.objects.filter(content_type_id=content_type.pk).values_list("object_id", flat=True)
            unversioned = Model.admin_manager.exclude(pk__in=version_ids).order_by("-pk")
            self.stdout.write(self.style.NOTICE(
                f"{len(version_ids) + len(unversioned)} objects of type {Model.__name__}, thereof "
                f"{len(unversioned)} missing Version object"
            ))
            if user is None and not options["dry_run"] and unversioned:  # pragma: no cover
                raise CommandError("Please specify a user which missing Version objects shall belong to "
                                   "either with the DJANGOCMS_VERSIONING_DEFAULT_USER setting or using "
                                   "command line arguments")

            for orphan in unversioned:
                # find all model instances that belong to the same grouper
                selectors = {versionable.grouper_field_name: getattr(orphan, versionable.grouper_field_name)}
                for extra_selector in versionable.extra_grouping_fields:
                    selectors[extra_selector] = getattr(orphan, extra_selector)
                same_grouper_ids = Model.admin_manager.filter(**selectors).values_list("pk", flat=True)
                # get all existing version objects
                existing_versions = Version.objects.filter(content_type=content_type, object_id__in=same_grouper_ids)
                # target state
                state = options["state"]
                # change to "archived" if state already exists
                if state != constants.ARCHIVED:
                    for version in existing_versions:
                        if version.state == state:
                            state = constants.ARCHIVED
                            break

                if options["dry_run"]:  # pragma: no cover
                    # Only write out change
                    self.stdout.write(self.style.NOTICE(
                        f"{str(orphan)} (pk={orphan.pk}) would be assigned a Version object with state {state}"
                    ))
                else:
                    try:
                        Version.objects.create(
                            content=orphan,
                            state=state,
                            created_by=user,
                        )
                        self.stdout.write(self.style.SUCCESS(
                            f"Successfully created version object for {Model.__name__} with pk={orphan.pk}"
                        ))
                    except Exception as e:  # pragma: no cover
                        self.stdout.write(self.style.ERROR(
                            f"Failed creating version object for {Model.__name__} with pk={orphan.pk}: {e}"
                        ))



from cms.test_utils.testcases import CMSTestCase
from cms.utils.urlutils import admin_reverse

from djangocms_versioning.helpers import get_latest_admin_viewable_content
from djangocms_versioning.models import Version
from djangocms_versioning.test_utils.blogpost.admin import BlogContentAdmin
from djangocms_versioning.test_utils.blogpost.models import BlogContent
from djangocms_versioning.test_utils.factories import (
    BlogContentFactory,
    BlogPostFactory,
    BlogPostVersionFactory,
    PageFactory,
    PageVersionFactory,
)


class TestLatestAdminViewable(CMSTestCase):

    def setUp(self) -> None:
        """Creates a page, page content and a version object for the following tests"""
        self.page = PageFactory()
        self.version = PageVersionFactory(
            content__page=self.page,
            content__language="en",
        )

    def test_extra_grouping_fields(self):
        # Test 1: Try getting content w/o language grouping field => needs to fail
        self.assertRaises(ValueError, lambda: get_latest_admin_viewable_content(self.page))  # no language grouper

        # Test 2: Try getting content w/ langauge grouping field => needs to succeed
        content = get_latest_admin_viewable_content(self.page, language="en")  # OK
        self.assertEqual(content.versions.first(), self.version)

    def test_latest_admin_viewable_draft(self):
        # New page has draft version, nothing else: latest_admin_viewable_content is draft
        content = get_latest_admin_viewable_content(self.page, language="en")
        self.assertEqual(content.versions.first(), self.version)

    def test_latest_admin_viewable_archive(self):
        # First archive draft
        self.version.archive(user=self.get_superuser())
        # Archived version, nothing else: latest_admin_viewable_content is empty
        content = get_latest_admin_viewable_content(self.page, include_unpublished_archived=False, language="en")
        self.assertIsNone(content)
        # Archived version, nothing else: latest_admin_viewable_content is empty
        content = get_latest_admin_viewable_content(self.page, include_unpublished_archived=True, language="en")
        self.assertEqual(content.versions.first(), self.version)

    def test_latest_admin_viewable_published(self):
        # Now revert and publish => latest content is published
        self.version.archive(user=self.get_superuser())
        version2 = self.version.copy(created_by=self.get_superuser())
        version2.publish(user=self.get_superuser())
        # Published version is always viewable
        content = get_latest_admin_viewable_content(self.page, include_unpublished_archived=False, language="en")
        self.assertEqual(content.versions.first(), version2)
        # Published version is always viewable
        content = get_latest_admin_viewable_content(self.page, include_unpublished_archived=True, language="en")
        self.assertEqual(content.versions.first(), version2)

    def test_latest_admin_viewable_draft_on_top_of_published(self):
        # Now create a draft on top of published -> latest_admin_viewable content will be draft
        self.version.publish(user=self.get_superuser())
        version2 = self.version.copy(created_by=self.get_superuser())
        # Draft version is shadows published version
        content = get_latest_admin_viewable_content(self.page, include_unpublished_archived=False, language="en")
        self.assertEqual(content.versions.first(), version2)
        # Draft version is shadows published version
        content = get_latest_admin_viewable_content(self.page, include_unpublished_archived=True, language="en")
        self.assertEqual(content.versions.first(), version2)

    def test_latest_admin_viewable_archive_on_top_of_published(self):
        # Archive draft, with published version available
        self.version.publish(user=self.get_superuser())
        version2 = self.version.copy(created_by=self.get_superuser())
        version2.archive(user=self.get_superuser())
        # Published version now is the latest version
        content = get_latest_admin_viewable_content(self.page, include_unpublished_archived=False, language="en")
        self.assertEqual(content.versions.first(), self.version)
        # Published version now is the latest version even when including archived
        content = get_latest_admin_viewable_content(self.page, include_unpublished_archived=True, language="en")
        self.assertEqual(content.versions.first(), self.version)


class TestVersionState(CMSTestCase):
    def test_page_indicators(self):
        """The page content indicators render correctly"""
        page = PageFactory(node__depth=1)
        version1 = PageVersionFactory(
            content__page=page,
            content__language="en",
        )
        pk = version1.pk

        page_tree = admin_reverse("cms_pagecontent_get_tree")
        with self.login_user_context(self.get_superuser()):
            # New page has draft version, nothing else
            response = self.client.get(page_tree, {"language": "en"})
            self.assertNotContains(response, "cms-pagetree-node-state-empty")
            self.assertContains(response, "cms-pagetree-node-state-draft")
            self.assertNotContains(response, "cms-pagetree-node-state-published")
            self.assertNotContains(response, "cms-pagetree-node-state-dirty")
            self.assertNotContains(response, "cms-pagetree-node-state-unpublished")

            # Now archive
            response = self.client.post(admin_reverse("djangocms_versioning_pagecontentversion_archive",
                                                      args=(pk,)))
            self.assertEqual(response.status_code, 302)  # Sends a redirect
            # Is archived indicator? No draft indicator
            response = self.client.get(page_tree, {"language": "en"})
            self.assertContains(response, "cms-pagetree-node-state-archived")
            self.assertNotContains(response, "cms-pagetree-node-state-draft")

            # Now revert
            response = self.client.post(admin_reverse("djangocms_versioning_pagecontentversion_revert",
                                        args=(pk,)))
            self.assertEqual(response.status_code, 302)  # Sends a redirect
            # Is draft indicator? No archived indicator
            response = self.client.get(page_tree, {"language": "en"})
            self.assertContains(response, "cms-pagetree-node-state-draft")
            self.assertNotContains(response, "cms-pagetree-node-state-archived")
            # New draft was created, get new pk
            pk = Version.objects.filter_by_content_grouping_values(version1.content).order_by("-pk")[0].pk

            # Now publish
            response = self.client.post(admin_reverse("djangocms_versioning_pagecontentversion_publish",
                                        args=(pk,)))
            self.assertEqual(response.status_code, 302)  # Sends a redirect
            # Is published indicator? No draft indicator
            response = self.client.get(page_tree, {"language": "en"})
            self.assertContains(response, "cms-pagetree-node-state-published")
            self.assertNotContains(response, "cms-pagetree-node-state-draft")

            # Now unpublish
            response = self.client.post(admin_reverse("djangocms_versioning_pagecontentversion_unpublish",
                                        args=(pk,)))
            self.assertEqual(response.status_code, 302)  # Sends a redirect

            # Is unpublished indicator? No published indicator
            response = self.client.get(page_tree, {"language": "en"})
            self.assertContains(response, "cms-pagetree-node-state-unpublished")
            self.assertNotContains(response, "cms-pagetree-node-state-published")

            # Now revert
            response = self.client.post(admin_reverse("djangocms_versioning_pagecontentversion_revert",
                                        args=(pk,)))
            self.assertEqual(response.status_code, 302)  # Sends a redirect

            # Is draft indicator? No unpublished indicator
            response = self.client.get(page_tree, {"language": "en"})
            self.assertContains(response, "cms-pagetree-node-state-draft")
            self.assertNotContains(response, "cms-pagetree-node-state-unpublished")
            # New draft was created, get new pk
            pk = Version.objects.filter_by_content_grouping_values(version1.content).order_by("-pk")[0].pk

            # Now archive
            response = self.client.post(admin_reverse("djangocms_versioning_pagecontentversion_archive",
                                        args=(pk,)))
            self.assertEqual(response.status_code, 302)  # Sends a redirect

            # Is archived indicator? No draft indicator
            response = self.client.get(page_tree, {"language": "en"})
            self.assertContains(response, "cms-pagetree-node-state-archived")
            self.assertNotContains(response, "cms-pagetree-node-state-draft")

            # Now revert
            response = self.client.post(admin_reverse("djangocms_versioning_pagecontentversion_revert",
                                        args=(pk,)))
            self.assertEqual(response.status_code, 302)  # Sends a redirect

            # Is draft indicator? No unpublished indicator
            response = self.client.get(page_tree, {"language": "en"})
            self.assertContains(response, "cms-pagetree-node-state-draft")
            self.assertNotContains(response, "cms-pagetree-node-state-unpublished")
            # New draft was created, get new pk
            pk = Version.objects.filter_by_content_grouping_values(version1.content).order_by("-pk")[0].pk

            # Now publish again and then edit redirect to create a draft on top of published version
            response = self.client.post(admin_reverse("djangocms_versioning_pagecontentversion_publish",
                                                      args=(pk,)))
            self.assertEqual(response.status_code, 302)  # Sends a redirect
            response = self.client.post(admin_reverse("djangocms_versioning_pagecontentversion_edit_redirect",
                                                      args=(pk,)))
            self.assertEqual(response.status_code, 302)  # Sends a redirect

            # Is published indicator? No draft indicator
            response = self.client.get(page_tree, {"language": "en"})
            self.assertContains(response, "cms-pagetree-node-state-dirty")
            self.assertNotContains(response, "cms-pagetree-node-state-published")

    def test_mixin_facory_media(self):
        """Test if the IndicatorMixin imports required js and css"""
        from django.contrib import admin

        admin = BlogContentAdmin(BlogContent, admin.site)
        self.assertIn("cms.pagetree.css", str(admin.media))
        self.assertIn("indicators.js", str(admin.media))

    def test_mixin_factory(self):
        """The IndicatorMixin causes the indicators to be rendered"""
        blogpost = BlogPostFactory()
        content = BlogContentFactory(
            blogpost=blogpost
        )
        BlogPostVersionFactory(
            content=content,
        )

        changelist = admin_reverse("blogpost_blogcontent_changelist")
        with self.login_user_context(self.get_superuser()):
            # New page has draft version, nothing else
            response = self.client.get(changelist)
            # Status indicator available?
            self.assertContains(response, "cms-pagetree-node-state-draft")
            self.assertNotContains(response, "cms-pagetree-node-state-published")
            self.assertNotContains(response, "cms-pagetree-node-state-dirty")
            # CSS loaded?
            self.assertContains(response, "cms.pagetree.css"),
            # JS loadeD?
            self.assertContains(response, "indicators.js")



from cms.plugin_base import CMSPluginBase
from cms.plugin_pool import plugin_pool

from .models import Text


@plugin_pool.register_plugin
class SimpleTextPlugin(CMSPluginBase):
    model = Text
    name = "SimpleText"
    allow_children = False
    render_template = "text/text.html"



# -*- coding: utf-8 -*-
# Generated by Django 1.11.14 on 2018-08-03 13:32
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("djangocms_versioning", "0004_auto_20180730_1135")]

    operations = [migrations.RemoveField(model_name="version", name="label")]



from django.apps import apps
from django.db.models.base import Model


def _cms_extension():
    return apps.get_app_config("djangocms_versioning").cms_extension


def _to_model(model_or_obj):
    if isinstance(model_or_obj, Model):
        model_or_obj = model_or_obj.__class__
    return model_or_obj


def for_content(model_or_obj):
    """Get the registered VersionableItem instance for a content model or content model instance"""
    return _cms_extension().versionables_by_content[_to_model(model_or_obj)]


def for_grouper(model_or_obj):
    """Get the registered VersionableItem instance for a grouper model or grouper model instance"""
    return _cms_extension().versionables_by_grouper[_to_model(model_or_obj)]


def exists_for_content(model_or_obj):
    """Test for registered VersionableItem for a content model or content model instance"""
    return _to_model(model_or_obj) in _cms_extension().versionables_by_content


def exists_for_grouper(model_or_obj):
    """Test for registered VersionableItem for a grouper model or grouper model instance"""
    return _to_model(model_or_obj) in _cms_extension().versionables_by_grouper



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2018-08-30 22:38
from __future__ import unicode_literals

from django.apps import apps
from django.db import migrations


def make_operations():
    versioning_extension = apps.get_app_config("djangocms_versioning").cms_extension

    return [
        migrations.CreateModel(
            name=versionable.version_model_proxy.__name__,
            fields=[],
            options={"managed": False, "proxy": True},
            bases=("djangocms_versioning.version",),
        )
        for versionable in versioning_extension.versionables
    ]


class Migration(migrations.Migration):

    dependencies = [
        ("djangocms_versioning", "0009_cms_pagecontent_remove_unique_constraint")
    ]

    operations = make_operations()



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2018-10-05 14:01
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("djangocms_versioning", "0010_version_proxies")]

    operations = [
        migrations.AddField(
            model_name="version",
            name="number",
            field=models.CharField(max_length=11, null=True),
        )
    ]



from cms.extensions import PageContentExtensionAdmin
from django.contrib import admin

from .models import PollPageContentExtension


@admin.register(PollPageContentExtension)
class PollExtensionAdmin(PageContentExtensionAdmin):

    pass



# -*- coding: utf-8 -*-
# Generated by Django 1.11.17.dev20181001130818 on 2018-11-30 10:08
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [("djangocms_versioning", "0014_version_source")]

    operations = [
        migrations.AddField(
            model_name="version",
            name="modified",
            field=models.DateTimeField(default=django.utils.timezone.now, verbose_name="Modified"),
        )
    ]



import typing
from urllib.parse import urljoin

from cms.toolbar.utils import get_object_preview_url
from cms.utils import get_current_site
from django.conf import settings
from django.contrib.sites.models import Site
from django.utils.translation import gettext_lazy as _

from djangocms_versioning import models
from djangocms_versioning.helpers import send_email


def get_full_url(location: str, site: typing.Union[Site, None] = None) -> str:
    if not site:
        site = Site.objects.get_current()

    if getattr(settings, "USE_HTTPS", False):
        scheme = "https"
    else:
        scheme = "http"
    domain = f"{scheme}://{site.domain}"
    return urljoin(domain, location)


def notify_version_author_version_unlocked(version: models.Version, unlocking_user: settings.AUTH_USER_MODEL) -> int:
    # If the unlocking user is the current author, don't send a notification email
    if version.created_by == unlocking_user:
        return 0

    # If the users name is available use it, otherwise use their username
    username = unlocking_user.get_full_name() or unlocking_user.username

    site = get_current_site()
    recipients = [version.created_by.email]
    subject = "[Django CMS] ({site_name}) {title} - {description}".format(
        site_name=site.name,
        title=version.content,
        description=_("Unlocked"),
    )
    version_url = get_full_url(
        get_object_preview_url(version.content)
    )

    # Prepare and send the email
    template_context = {
        "version_link": version_url,
        "by_user": username,
    }
    status = send_email(
        recipients=recipients,
        subject=subject,
        template="unlock-notification.txt",
        template_context=template_context,
    )
    return status



from cms.plugin_base import CMSPluginBase
from cms.plugin_pool import plugin_pool

from .models import PollPlugin as Poll


@plugin_pool.register_plugin
class PollPlugin(CMSPluginBase):
    model = Poll
    name = "Poll"
    allow_children = True
    render_template = "polls/poll.html"
